from django.template import Context, Template
from django.test import SimpleTestCase
from django.utils.html import format_html

from laces.components import Component


class TestComponentTag(SimpleTestCase):
    """
    Test for the `component` template tag.

    Extracted from Wagtail. See:
    https://github.com/wagtail/wagtail/blob/main/wagtail/admin/tests/test_templatetags.py#L225-L305  # noqa: E501
    """

    def test_passing_context_to_component(self):
        class MyComponent(Component):
            def render_html(self, parent_context):
                return format_html(
                    "<h1>{} was here</h1>", parent_context.get("first_name", "nobody")
                )

        template = Template(
            "{% load laces %}{% with first_name='Kilroy' %}{% component my_component %}{% endwith %}"
        )
        html = template.render(Context({"my_component": MyComponent()}))
        self.assertEqual(html, "<h1>Kilroy was here</h1>")

        template = Template(
            "{% load laces %}{% component my_component with first_name='Kilroy' %}"
        )
        html = template.render(Context({"my_component": MyComponent()}))
        self.assertEqual(html, "<h1>Kilroy was here</h1>")

        template = Template(
            "{% load laces %}{% with first_name='Kilroy' %}{% component my_component with surname='Silk' only %}{% endwith %}"
        )
        html = template.render(Context({"my_component": MyComponent()}))
        self.assertEqual(html, "<h1>nobody was here</h1>")

    def test_fallback_render_method(self):
        class MyComponent(Component):
            def render_html(self, parent_context):
                return format_html("<h1>I am a component</h1>")

        class MyNonComponent:
            def render(self):
                return format_html("<h1>I am not a component</h1>")

        template = Template("{% load laces %}{% component my_component %}")
        html = template.render(Context({"my_component": MyComponent()}))
        self.assertEqual(html, "<h1>I am a component</h1>")
        with self.assertRaises(ValueError):
            template.render(Context({"my_component": MyNonComponent()}))

        template = Template(
            "{% load laces %}{% component my_component fallback_render_method=True %}"
        )
        html = template.render(Context({"my_component": MyComponent()}))
        self.assertEqual(html, "<h1>I am a component</h1>")
        html = template.render(Context({"my_component": MyNonComponent()}))
        self.assertEqual(html, "<h1>I am not a component</h1>")

    def test_component_escapes_unsafe_strings(self):
        class MyComponent(Component):
            def render_html(self, parent_context):
                return "Look, I'm running with scissors! 8< 8< 8<"

        template = Template("{% load laces %}<h1>{% component my_component %}</h1>")
        html = template.render(Context({"my_component": MyComponent()}))
        self.assertEqual(
            html, "<h1>Look, I&#x27;m running with scissors! 8&lt; 8&lt; 8&lt;</h1>"
        )

    def test_error_on_rendering_non_component(self):
        template = Template("{% load laces %}<h1>{% component my_component %}</h1>")

        with self.assertRaises(ValueError) as cm:
            template.render(Context({"my_component": "hello"}))
        self.assertEqual(str(cm.exception), "Cannot render 'hello' as a component")

    def test_render_as_var(self):
        class MyComponent(Component):
            def render_html(self, parent_context):
                return format_html("<h1>I am a component</h1>")

        template = Template(
            "{% load laces %}{% component my_component as my_html %}The result was: {{ my_html }}"
        )
        html = template.render(Context({"my_component": MyComponent()}))
        self.assertEqual(html, "The result was: <h1>I am a component</h1>")



from typing import Any, MutableMapping

from django.forms import Media, MediaDefiningClass
from django.template import Context
from django.template.loader import get_template


class Component(metaclass=MediaDefiningClass):
    """
    A class that knows how to render itself.

    Extracted from Wagtail. See:
    https://github.com/wagtail/wagtail/blob/094834909d5c4b48517fd2703eb1f6d386572ffa/wagtail/admin/ui/components.py#L8-L22  # noqa: E501
    """

    def get_context_data(
        self, parent_context: MutableMapping[str, Any]
    ) -> MutableMapping[str, Any]:
        return {}

    def render_html(self, parent_context: MutableMapping[str, Any] = None) -> str:
        """
        Return string representation of the object.

        Given a context dictionary from the calling template (which may be a
        `django.template.Context` object or a plain ``dict`` of context variables),
        returns the string representation to be rendered.

        This will be subject to Django's HTML escaping rules, so a return value
        consisting of HTML should typically be returned as a
        `django.utils.safestring.SafeString` instance.
        """
        if parent_context is None:
            parent_context = Context()
        context_data = self.get_context_data(parent_context)
        if context_data is None:
            raise TypeError("Expected a dict from get_context_data, got None")

        template = get_template(self.template_name)
        return template.render(context_data)


class MediaContainer(list):
    """
    A list that provides a ``media`` property that combines the media definitions
    of its members.

    Extracted from Wagtail. See:
    https://github.com/wagtail/wagtail/blob/ca8a87077b82e20397e5a5b80154d923995e6ca9/wagtail/admin/ui/components.py#L25-L36  # noqa: E501
    """

    @property
    def media(self):
        media = Media()
        for item in self:
            media += item.media
        return media



from django import template
from django.template.base import token_kwargs
from django.template.defaultfilters import conditional_escape


register = template.library.Library()


class ComponentNode(template.Node):
    """
    Template node to render a component.

    Extracted from Wagtail. See:
    https://github.com/wagtail/wagtail/blob/094834909d5c4b48517fd2703eb1f6d386572ffa/wagtail/admin/templatetags/wagtailadmin_tags.py#L937-L987  # noqa: E501
    """

    def __init__(
        self,
        component,
        extra_context=None,
        isolated_context=False,
        fallback_render_method=None,
        target_var=None,
    ):
        self.component = component
        self.extra_context = extra_context or {}
        self.isolated_context = isolated_context
        self.fallback_render_method = fallback_render_method
        self.target_var = target_var

    def render(self, context: template.Context) -> str:
        # Render a component by calling its render_html method, passing request and context from the
        # calling template.
        # If fallback_render_method is true, objects without a render_html method will have render()
        # called instead (with no arguments) - this is to provide deprecation path for things that have
        # been newly upgraded to use the component pattern.

        component = self.component.resolve(context)

        if self.fallback_render_method:
            fallback_render_method = self.fallback_render_method.resolve(context)
        else:
            fallback_render_method = False

        values = {
            name: var.resolve(context) for name, var in self.extra_context.items()
        }

        if hasattr(component, "render_html"):
            if self.isolated_context:
                html = component.render_html(context.new(values))
            else:
                with context.push(**values):
                    html = component.render_html(context)
        elif fallback_render_method and hasattr(component, "render"):
            html = component.render()
        else:
            raise ValueError(f"Cannot render {component!r} as a component")

        if self.target_var:
            context[self.target_var] = html
            return ""
        else:
            if context.autoescape:
                html = conditional_escape(html)
            return html


@register.tag(name="component")
def component(parser, token):
    """
    Template tag to render a component via ComponentNode.

    Extracted from Wagtail. See:
    https://github.com/wagtail/wagtail/blob/094834909d5c4b48517fd2703eb1f6d386572ffa/wagtail/admin/templatetags/wagtailadmin_tags.py#L990-L1037  # noqa: E501
    """
    bits = token.split_contents()[1:]
    if not bits:
        raise template.TemplateSyntaxError(
            "'component' tag requires at least one argument, the component object"
        )

    component = parser.compile_filter(bits.pop(0))

    # the only valid keyword argument immediately following the component
    # is fallback_render_method
    flags = token_kwargs(bits, parser)
    fallback_render_method = flags.pop("fallback_render_method", None)
    if flags:
        raise template.TemplateSyntaxError(
            "'component' tag only accepts 'fallback_render_method' as a keyword argument"
        )

    extra_context = {}
    isolated_context = False
    target_var = None

    while bits:
        bit = bits.pop(0)
        if bit == "with":
            extra_context = token_kwargs(bits, parser)
        elif bit == "only":
            isolated_context = True
        elif bit == "as":
            try:
                target_var = bits.pop(0)
            except IndexError:
                raise template.TemplateSyntaxError(
                    "'component' tag with 'as' must be followed by a variable name"
                )
        else:
            raise template.TemplateSyntaxError(
                "'component' tag received an unknown argument: %r" % bit
            )

    return ComponentNode(
        component,
        extra_context=extra_context,
        isolated_context=isolated_context,
        fallback_render_method=fallback_render_method,
        target_var=target_var,
    )
